/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * jtalk_rl.c: readline-based front-end
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <locale.h>
#include <poll.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <readline/readline.h>

/*
   esc 7 -> esc 8
   esc s ? -> esc u ?
*/

#define SAVE_CURSOR "\0337"
#define RESTORE_CURSOR_AND_CLEAR "\0338\033[J"

#define RL_PROMPT_START_IGNORE_S "\001"
#define RL_PROMPT_END_IGNORE_S "\002"

#define SAVE_CURSOR_PROMPT RL_PROMPT_START_IGNORE_S SAVE_CURSOR \
    RL_PROMPT_END_IGNORE_S

static int sock;
static unsigned char *buffer;
static size_t buffer_size, buffer_have;

static void
display_message(const char *t)
{
    printf(RESTORE_CURSOR_AND_CLEAR "%s\033[m\n", t);
    rl_forced_update_display();
}

static void
send_to_jtalkd(const char *t)
{
    size_t l;
    ssize_t w;

    l = strlen(t) + 1;
    while(l > 0) {
	if((w = write(sock, t, l)) < 0) {
	    if(errno == EINTR)
		continue;
	    perror("write");
	    exit(1);
	}
	l -= w;
	t += w;
    }
}

static void
have_keyboard_line(char *line)
{
    printf(RESTORE_CURSOR_AND_CLEAR);
    if(line == NULL)
	return;
    send_to_jtalkd(line);
    free(line);
}

static void
have_network_line(char *line)
{
    display_message(line);
}

static void
set_socket_path(char *p, size_t lp)
{
    char *e;

    if((e = getenv("JTALK_BASE")) != NULL) {
	snprintf(p, lp, "%s/socket/s", e);
    } else {
	if((e = getenv("HOME")) == NULL) {
	    fprintf(stderr, "No $HOME nor $JTALK_BASE.\n");
	    exit(1);
	}
	snprintf(p, lp, "%s/.jtalk/socket/s", e);
    }
}

static void
connect_jtalkd(void)
{
    struct sockaddr_un addr;

    if((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	exit(1);
    }
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    set_socket_path(addr.sun_path, sizeof(addr.sun_path));
    if(connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	perror("connect");
	exit(1);
    }
}

static void
read_from_jtalkd(void)
{
    ssize_t l;
    unsigned char *end;

    if(buffer_size - buffer_have < 1024) {
	buffer_size = buffer_size == 0 ? 4096 : buffer_size * 2;
	if((buffer = realloc(buffer, buffer_size)) == NULL) {
	    perror("malloc");
	    exit(1);
	}
    }
    if((l = read(sock, buffer + buffer_have, buffer_size - buffer_have)) < 0) {
	if(errno == EINTR)
	    return;
	perror("read");
	exit(1);
    }
    if(l == 0)
	exit(0);
    buffer_have += l;
    /* TODO: make that a little bit faster: do not re-search what is already
     * here, do not memmove for all lines (not really important, most of the
     * time I will have exactly one line at once in the buffer) */
    while((end = memchr(buffer, 0, buffer_have)) != NULL) {
	have_network_line(buffer);
	end++;
	buffer_have -= end - buffer;
	memmove(buffer, end, buffer_have);
    }
}

static void
send_argv(int argc, char **argv)
{
    size_t l;

    while(argc > 0) {
	l = strlen(*argv);
	char b[7 + l];
	sprintf(b, "/argv %s", *argv);
	send_to_jtalkd(b);
	argv++;
	argc--;
    }
    send_to_jtalkd("/argv");
}

int
main(int argc, char **argv)
{
    struct pollfd fds[2];

    setlocale(LC_CTYPE, "");
    printf("\033%%G");
    buffer = NULL;
    buffer_size = 0;
    buffer_have = 0;
    connect_jtalkd();
    send_argv(argc - 1, argv + 1);
    rl_callback_handler_install(SAVE_CURSOR_PROMPT ">", have_keyboard_line);
    fds[0].fd = 0;
    fds[0].events = POLLIN;
    fds[1].fd = sock;
    fds[1].events = POLLIN;
    while(1) {
	if(poll(fds, 2, -1) < 0) {
	    if(errno == EINTR)
		continue;
	    perror("poll");
	    exit(1);
	}
	if((fds[0].revents & POLLIN) != 0)
	    rl_callback_read_char();
	if((fds[1].revents & POLLIN) != 0)
	    read_from_jtalkd();
    }
    return(0);
}
